#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

# Generate Config

require 'yaml'
require 'uri'
require 'fileutils'

DEBUG = true

class GenerateConfig

  private

  def initialize(cmd, argv)
    @cmd = cmd
    @argv = argv
  end

  # 引数の数のチェックとusageの出力
  # argv.size == 3 は隠しオプション（パスの指定）
  def check_args_num
    if @argv.size < 1 or @argv.size > 3 then
      STDERR.puts "usage: #{@cmd} <SIDfm VM Server URL> <hostname:optional>"
      STDERR.puts "  ex1. #{@cmd} http://sidfmvm.example.co.jp/redmine/sidfm/api/register/ host1.example.com"
      STDERR.puts "  ex2. #{@cmd} http://sidfmvm.example.co.jp/redmine/sidfm/api/register/"
      exit false
    end
  end

  # URLのチェック
  # 末尾に / がないURLは、強制的に付与する。
  def check_url
    @url = nil
    begin
      path = @argv[0]
      if '/' != path[path.size - 1]  then
        path += '/'
      end
      @url = URI.parse(path)
      if !@url.kind_of?(URI::HTTP) and !@url.kind_of?(URI::HTTPS) then
        STDERR.puts "invalid url - http/https url only: #{@url}"
      end
    rescue Exception => e
      STDERR.puts e.message
      e.backtrace.each do |i|
        STDERR.puts i
      end
      STDERR.puts "invalid url - parse error: #{@url}"
      exit false
    end
  end

  # ホスト名のチェック
  def check_hostname
    @hostname = nil
    return if @argv.size < 2
    @hostname = @argv[1]
    if @hostname.size == 0 or @hostname !~ /^[0-9a-z\.-]+$/i then
      STDERR.puts "invalid hostname: #{@hostname}"
      exit false
    end
  end

  # base_pathのチェック
  def check_base_path
    @base_path = '/opt/softek/'
    return if @argv.size < 3
    @base_path = @argv[2]
    @base_path += '/' if @base_path[@base_path.size - 1] != '/'
    if @base_path.size == 0 then
      STDERR.puts "invalid base_path: #{@base_path}"
      exit false
    end
  end

  # 引数チェック
  def check_args
    check_args_num
    check_url
    check_hostname
    check_base_path
  end

  # ディストリビューションのチェック
  def check_env
    @env = FileTest.exist?('/etc/debian_version') ? 'debian' : 'centos'
    @config_filename = "#{@env}_agent.yaml"
  end

  # コンフィグファイルが既に存在していた場合には、ファイル名を変更して対比。
  def backup_old_file
    if FileTest.exist?(@base_path + 'bin/' + @config_filename) then
      FileUtils.mv(@base_path + 'bin/' + @config_filename, @base_path + 'bin/' + @config_filename + '.old', :force => true)
    end
  end

  # 設定ファイル作成
  def create_config
    YAML.dump({
                :api => {
                  :api_url => @url.to_s,
                  :hostname => @hostname ? @hostname : '',
                  :base_path => @base_path,
                  :bin_path => 'bin/',
                  :register_apikey_path => 'tmp/.register_apikey',
                  :host_apikey_path => 'var/.host_apikey',
                  :hostname_path => 'var/.hostname',
                },
              },
              File.open(@base_path + 'bin/' + @config_filename, 'w'))
  end

  public

  def run
    check_args
    check_env
    backup_old_file
    create_config
  end

end

if __FILE__ == $0 then
  id = GenerateConfig.new($0, ARGV)
  id.run
end
