#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

# Proxy Config
# Copyright (C) 2019- SofTek Systems, Inc. All rights reserved.

require 'yaml'
require 'uri'
require 'fileutils'

DEBUG = true

class ProxyConfig

  BASE_PATH = '/opt/softek/'
  
  private

  def initialize(cmd, argv)
    @cmd = cmd
    @argv = argv
    @enable = false
    @http = nil
    @https = nil
    @base_path = BASE_PATH
  end

  # 引数の数のチェックとusageの出力
  def check_args_num
    if @argv.size < 1 or @argv.size > 3 then
      STDERR.puts "usage: #{@cmd} <enable|disable> <http proxy> <https proxy>"
      STDERR.puts "  ex."
      STDERR.puts "    proxy disable: #{@cmd} disable"
      STDERR.puts "    proxy enable1: #{@cmd} enable proxy.example.com:8080"
      STDERR.puts "    proxy enable2: #{@cmd} enable http.proxy.example.com:8080 https.proxy.example.com:8080"
      STDERR.puts "    proxy enable3: #{@cmd} enable user:pass@proxy.example.com:8080"
      exit false
    end
  end

  # enable/disableのチェック
  def check_enable
    if /\A(enable|disable)\z/i =~ @argv[0] then
      keyword = Regexp.last_match(1)
      if /\Aenable\z/i =~ keyword then
        @enable = true
      end
    else
      STDERR.puts "invalid argument - first arg must be 'enable' or 'disable'"
      exit false
    end
  end

  PROXY_PATTERN = (/\A(([^:]+)(:([^:]+))?@)?([0-9a-z.-]+):(\d+)\z/i).freeze
  
  def check_http_proxy
    return unless @enable
    if @argv.size < 2 then
      STDERR.puts "invalid argument - lack http proxy"
      exit false
    end
    if PROXY_PATTERN =~ @argv[1] then
      @http = {
        :user => Regexp.last_match(2),
        :pass => Regexp.last_match(4),
        :host => Regexp.last_match(5),
        :port => Regexp.last_match(6).to_i,
      }
    else
      STDERR.puts "invalid argument - invalid http proxy"
      exit false
    end
  end

  def check_https_proxy
    return unless @enable
    if @argv.size < 3 then
      return
    end
    if PROXY_PATTERN =~ @argv[2] then
      @https = {
        :user => Regexp.last_match(2),
        :pass => Regexp.last_match(4),
        :host => Regexp.last_match(5),
        :port => Regexp.last_match(6).to_i,
      }
    else
      STDERR.puts "invalid argument - invalid https proxy"
      exit false
    end
  end

  # 引数チェック
  def check_args
    check_args_num
    check_enable
    check_http_proxy
    check_https_proxy
   end

  def check_env
    @config_filename = "proxy_config.yaml"
  end

  # コンフィグファイルが既に存在していた場合には、ファイル名を変更して対比。
  def backup_old_file
    if FileTest.exist?(@base_path + 'bin/' + @config_filename) then
      FileUtils.mv(@base_path + 'bin/' + @config_filename, @base_path + 'bin/' + @config_filename + '.old', :force => true)
    end
  end

  # 設定ファイル作成
  def create_config
    YAML.dump({
                :proxy => {
                  :enable => @enable,
                  :http => @http,
                  :https => @https,
                },
              },
              File.open(@base_path + 'bin/' + @config_filename, 'w'))
  end
  
  public

  def run
    check_args
    check_env
    backup_old_file
    create_config
  end
  
end

if __FILE__ == $0 then
  id = ProxyConfig.new($0, ARGV)
  id.run
end
