#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

require 'fileutils'
require 'socket'
require 'uri'
require 'openssl'
require 'net/http'
require 'yaml'
require File.expand_path(File.join(File.dirname(__FILE__), 'simple-json-parser-builder.rb'))

DEBUG = false

# SidfmVMAgentから登録（ホスト、データ）用APIへのアクセスを提供するクラス
class SidfmVMRegisterAPI

  private

  def config_path(sym)
    base_path = @config[:base_path] || ''
    base_path + (@config[sym] || '')
  end

  # register_apikeyもしくはhost_apikeyは必須。
  def initialize(config = {})
    raise ArgumentError, 'api_url is required' unless config[:api_url]

    @json_parser = JsonParser.new
    @json_builder = JsonBuilder.new
    @config = config
    @api_url = URI.parse(@config[:api_url])

    if @config[:register_apikey_path] and
        FileTest.readable?(config_path(:register_apikey_path)) then
      open(config_path(:register_apikey_path), 'r') do |f|
        @config[:register_apikey] = f.read.chomp
      end
    end

    if @config[:hostname].nil? or @config[:hostname].size == 0 then
      @config[:hostname] = Socket.gethostname
    end

    if @config[:host_apikey_path] and
        FileTest.readable?(config_path(:host_apikey_path)) then
      open(config_path(:host_apikey_path), 'r') do |f|
        @config[:host_apikey] = f.read.chomp
      end
    end

    # @config[:hostname_path]が無かったら、標準のパスを設定
    if !@config[:hostname_path] then
      @config[:hostname_path] = 'var/.hostname'
    end
    
    # hostname_pathにファイルがあったら、ホスト名を上書きする。
    if FileTest.readable?(config_path(:hostname_path)) then
      # 設定ファイルにhostname_pathがあれば、そちらを参照
      open(config_path(:hostname_path), 'r') do |f|
        @config[:hostname] = f.read.chomp
      end
    end

    @proxy_config = nil
    if FileTest.readable?(config_path(:bin_path)) then
      begin
        File.open(config_path(:bin_path) + 'proxy_config.yaml', 'r') do |f|
          @proxy_config = YAML.load(f.read)
        end
      rescue Exception => e
      end
    end
  end

  def proxy(is_https = false)
    ret = Net::HTTP::Proxy(nil)
    if @proxy_config then
      if @proxy_config[:proxy] then
        if @proxy_config[:proxy][:enable] then
          http = @proxy_config[:proxy][:http]
          https = @proxy_config[:proxy][:https]
          if https.nil? then
            ret = Net::HTTP::Proxy(http[:host], http[:port], http[:user], http[:pass])
          else
            if is_https then
              ret = Net::HTTP::Proxy(https[:host], https[:port], https[:user], https[:pass])
            else
              ret = Net::HTTP::Proxy(http[:host], http[:port], http[:user], http[:pass])
            end
          end
        end
      end
    end
    ret
  end

  def set_ssl(con)
    con.use_ssl = true
    con.verify_mode = OpenSSL::SSL::VERIFY_PEER
    ca_file = File.dirname(__FILE__) + '/sidfm_vm_ca.crt'
    if FileTest.readable?(ca_file) then
      con.ca_file = ca_file
    end
    con.set_debug_output(STDERR) if DEBUG
  end
  
  def get_request(path = '')
    ret = nil
    begin
      is_https = (@api_url.scheme == 'https')
      p = proxy(is_https)
      con = p.new(@api_url.host, @api_url.port)
      set_ssl(con) if is_https
      res = con.get(@api_url.path + path)
      body = res.body
      raise StandardError, 'null response:' + res.inspect unless body
      ret = @json_parser.parse(body)
    rescue Exception => e
      if DEBUG then
        STDERR.puts e.message
        e.backtrace.each do |i|
          STDERR.puts i
        end
      end
    end
    ret
  end

  def post_request(path = '', form_data = {}, headers = {})
    ret = nil
    begin
      is_https = (@api_url.scheme == 'https')
      p = proxy(is_https)
      con = p.new(@api_url.host, @api_url.port)
      set_ssl(con) if is_https
      url = @api_url + path
      req = Net::HTTP::Post.new(url.path, headers)
      if form_data.kind_of?(Hash) or form_data.kind_of?(Array) then
        req.set_form_data(form_data)
      else
        req.body = form_data
      end
      res = con.request(req)
      body = res.body
      raise StandardError, 'null response:' + res.inspect unless body
      ret = @json_parser.parse(body)
    rescue Exception => e
      if DEBUG then
        STDERR.puts e.message
        e.backtrace.each do |i|
          STDERR.puts i
        end
      end
    end
    ret
  end

  public

  def reachable?
    get_request('index')
  end

  def register_host
    ret = { :success => false }
    if @config[:host_apikey].nil? or @config[:host_apikey].size == 0 then
      ret = post_request('register_host', {
                           :'params[hostname]' => @config[:hostname],
                           :'params[register_apikey]' => @config[:register_apikey]
                         })
      if ret.respond_to?(:[]) then
        ret[:success] = ret['success']
        ret.delete('success')
        ret[:host_apikey] = ret['host_apikey']
        ret.delete('host_apikey')
        ret[:hostname] = ret['hostname']
        ret.delete('hostname')
        ret[:reason_id] = ret['reason_id']
        ret.delete('reason_id')
        ret[:reason] = ret['reason']
        ret.delete('reason')
        if ret[:success] then
          @config[:host_apikey] = ret[:host_apikey]
          @config[:hostname] = ret[:hostname]
          # 取得したホストAPIキーを保存
          if @config[:host_apikey_path] then
            File.open(config_path(:host_apikey_path), 'w') do |f|
              f.write(@config[:host_apikey])
            end
          end
          # 登録用APIキーを削除
          if @config[:register_apikey_path] then
            begin
              FileUtils.rm(config_path(:register_apikey_path), :force => true)
            rescue
              STDERR.puts 'cannot delete a file: ' + config_path(:register_apikey_path)
            end
          end
          # hostnameを保存
          if @config[:hostname_path] then
            File.open(config_path(:hostname_path), 'w') do |f|
              f.write(@config[:hostname])
            end
          end
        else
          if ret[:reason_id] then
            STDERR.puts "#{ret[:reason_id]}:#{ret[:reason]}"
            STDERR.puts ret.inspect
          else
            STDERR.puts 'Unpredictable error'
            STDERR.puts ret.inspect
          end
        end
      else
        STDERR.puts 'Unpredictable error'
        STDERR.puts ret.inspect
      end
    else
      ret[:success] = true
    end
    ret
  end

  def send_client_informations(data = {})
    data[:hostname] = @config[:hostname]
    data[:host_apikey] = @config[:host_apikey]
    ret = post_request('send_client_informations', @json_builder.build(data))
    if ret.respond_to?(:[]) and ret['success'] then
      ret[:success] = ret['success']
      ret.delete('success')
    end
    ret
  end

end

if __FILE__ == $0 then


  yaml = {}
  File.open('./centos_agent.yaml', 'r') do |f|
    yaml = YAML.load(f.read)
  end

  o = SidfmVMRegisterAPI.new(yaml[:api])
  p o.reachable?
  #p o.register_host
  #p o.send_client_informations({ :test => 'test1' })

end
