#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

require 'open3'
require 'yaml'
require 'socket'
require File.expand_path(File.join(File.dirname(__FILE__), 'simple-json-parser-builder.rb'))

require File.expand_path(File.join(File.dirname(__FILE__), 'sidfm_vm_register_api.rb'))

# SidfmVMUnixエージェント
#   Linuxを始めとしたUnix系OS用のSIDfmVMエージェント
class SidfmVMUnixAgent

  private

  def initialize(config = nil, cmdset = {}, applications = {})
    raise ArgumentError, 'null config' unless config
    raise ArgumentError, 'unreadable config: ' + config if !FileTest.readable?(config)
    @json_parser = JsonParser.new
    @json_builder = JsonBuilder.new
    @config = nil
    open(config, 'r') do |f|
      @config = YAML.load(f)
    end
    @cmdset = cmdset
    @applications = applications

    # Stringのインスタンスにforce_encodingメソッドがあるかどうかを
    # 確認して保存. ruby 1.8でも動作するように。
    @enable_force_encoding = String.new.respond_to?(:force_encoding)

    @api = SidfmVMRegisterAPI.new(@config[:api])
  end

  public

  # strのエンコーディングをASCII-8BITに設定
  def to_ascii8bit(str)
    @enable_force_encoding ? str.force_encoding('ascii-8bit') : str
  end

  # 指定したコマンドを実行して、標準出力を文字列として返す。
  def exec_command_stdout(command)
    ret = ''
    Open3.popen3(command) do |stdin, stdout, stderr, t|
      stdin.close_write
      ret = stdout.read
      #stderr.read
      t.value if t
    end
    ret
  end

  # 指定したコマンドを実行して、標準エラー出力を文字列として返す。
  def exec_command_stderr(command)
    ret = ''
    Open3.popen3(command) do |stdin, stdout, stderr, t|
      stdin.close_write
      ret = stderr.read
      #stdout.read
      t.value if t
    end
    ret
  end

  # 指定したコマンドを実行して、終了ステータスを返す。
  def exec_command_status(command)
    `#{command} 2>&1 >/dev/null` ; $?
  end

  # findコマンドの引数を作成する
  # 以下の形式のデータを渡すと、find用の引数を文字列として返す。
  # [ { :pattern => '正規表現パターン', :type => 'f' }, ... ]
  # fstypeは特定のファイルシステムのみを探索対象とする設定。
  def create_find_options(patterns = [], fstypes = [])
    find_opts = []
    patterns.each do |pattern|
      app_opts = []
      if pattern[:pattern] and pattern[:type] then
        app_opts << "\\( -regex '#{pattern[:pattern]}' -type #{pattern[:type]} \\)"
      end
      find_opts << app_opts.join(' -o ') if app_opts.size > 0
    end
    ret = find_opts.join(' -o ')
    if fstypes.size > 0 then
      fstype_opts = []
      fstypes.each do |fstype|
        fstype_opts << "! -fstype #{fstype}"
      end
      ret = "\\( #{fstype_opts.join(' ')} \\) -prune -o \\( #{ret} \\)"
    end
    ret
  end

  # findコマンドで@applicationsに該当するファイルパスを列挙する。
  def find_application_files
    patterns = []
    @applications.each_key do |x|
      y = @applications[x][:find_patterns]
      patterns << (y.size > 0 ? y : nil)
    end
    patterns.compact!
    find_opts = create_find_options(patterns.flatten, [ 'xfs', 'ext4' ])
    find_paths = exec_command_stdout("#{@cmdset[:find][:cmd]} / #{find_opts}").split("\n")
    app_paths = []
    find_paths.each do |find_path|
      status = check_package_by_path(find_path)
      if status and status != 0 then
        app_paths << find_path
      elsif DEBUG and find_path =~ /PHP/i then # FOR DEBUG
        app_paths << find_path                 # FOR DEBUG
      end
    end
    app_paths
  end

  # find_application_filesで取得した各ファイルパスについて、
  # 実際に対象アプリケーションであることのチェック及びバージョン取得
  # を行う。
  def analyze_application_paths(app_paths)
    ret = []
    @applications.each_key do |app|
      filtered_paths = []
      app_paths.each do |path|
        filtered_paths = []
        @applications[app][:find_patterns].each do |find_pattern|
          if Regexp.new(find_pattern[:pattern]) =~ path then
            filtered_paths << path
            break
          end
        end
        filtered_paths.uniq!
        if filtered_paths.size > 0 then
          filtered_paths.each do |filtered_path|
            checked = @applications[app][:check_proc].call(self, filtered_path)
            if checked[:success] then
              ret << [ checked[:application], checked[:version], filtered_path ]
            end
          end
        end
      end
    end
    ret
  end

  # cmdsetのコマンドがインストール済みかどうかを確認し、インストール
  # されていないものがあればエラーを配列で返す。
  # 全てインストール済なら空の配列を返す。
  def check_required_packages
    ret = []
    @cmdset.each_key do |cmd|
      checks = @cmdset[cmd][:checks]
      checks.each do |check|
        if FileTest.executable?(check) then
          @cmdset[cmd][:cmd] = check
          break
        end
      end
      unless @cmdset[cmd][:cmd] then
        ret << { :cmd => cmd, :cmd_info => @cmdset[cmd], :message => 'require package: ' + @cmdset[cmd][:package] }
      end
    end
    ret
  end

  # RHEL系ディストリビューションのバージョンを確認する。
  # サポート外のものも対応。
  # deprecated. get_os_releaseを使うことを推奨。
  RHEL_RELEASES = [ '/etc/centos-release',
                    '/etc/redhat-release',
                    '/etc/fedora-release',
                    '/etc/system-release' ]
  def get_rhel_release
    releases = []
    RHEL_RELEASES.each do |r|
      if FileTest.readable?(r) then
        File.open(r, 'r') do |f|
          releases << f.gets.chomp
        end
      end
    end
    ret = nil
    releases.each do |r|
      if /CentOS Linux release (\d+\.\d+)/ =~ r then
        ret = { :id => 'centos', :os => 'CentOS', :version => Regexp.last_match(1) }
        break
      elsif /CentOS release (\d+\.\d+)/ =~ r then
        ret = { :id => 'centos', :os => 'CentOS', :version => Regexp.last_match(1) }
        break
      elsif /Red Hat Enterprise Linux Server release (\d+\.\d+)/ =~ r then
        ret = { :id => 'rhel', :os => 'Red Hat Enterprise Linux', :version => Regexp.last_match(1) }
        break
      elsif /Amazon Linux 2/ =~ r then
        ret = { :id => 'amzn', :os => 'Amazon Linux', :version => '2' }
        break
      elsif /Amazon Linux AMI release (\d+\.\d+)/ =~ r then
        ret = { :id => 'amzn', :os => 'Amazon Linux', :version => '1' }
        break
      elsif /Fedora release (\d+)/ =~ r then
        ret = { :id => 'fedora', :os => 'Fedora', :version => Regexp.last_match(1) }
        break
      end
    end
    ret
  end

  # /etc/os-releaseから、ディストリビューション名とバージョンを取得
  OS_RELEASE = '/etc/os-release'

  def get_os_release
    ret = nil
    if FileTest.readable?(OS_RELEASE) then
      name_version = {}
      File.open(OS_RELEASE, 'r') do |f|
        while l = f.gets do
          l = l.chomp
          if /^ID="?([^"]+)"?$/ =~ l then
            name_version[:name] = Regexp.last_match(1)
            break if name_version[:name] and name_version[:version]
          elsif /^VERSION_ID="?([^"]+)"?$/ =~ l then
            name_version[:version] = Regexp.last_match(1)
            break if name_version[:name] and name_version[:version]
          end
        end
      end
      if name_version[:name] and name_version[:version] then
        ret = { :id => name_version[:name], :version => name_version[:version] }
        case name_version[:name]
        when 'amzn' then
          ret[:os] = 'Amazon Linux'
        when 'rhel' then
          ret[:os] = 'Red Hat Enterprise Linux'
        when 'centos' then
          ret[:os] = 'CentOS'
        when 'fedora' then
          ret[:os] = 'Fedora'
        when 'debian' then
          ret[:os] = 'Debian'
        when 'ubuntu' then
          ret[:os] = 'Ubuntu'
        when 'suse' then
          ret[:os] = 'SUSE'
        else
          ret[:os] = 'unknown'
        end
      end
    end
    ret
  end

  def register
    if @api.reachable? then
      @api.register_host
    else
      STDERR.puts @api.inspect if DEBUG
      raise RuntimeError, 'cannot access api'
    end
  end

  def send_data
    if @api.reachable? then
      data = {
        :os => guess_os,
        :packages => get_package_informations,
        :applications => get_application_versions
      }
      @api.send_client_informations(data)
    else
      STDERR.puts @api.inspect if DEBUG
      raise RuntimeError, 'cannot access api'
    end
  end

  def save_data(out_file, hostname = nil)
    unless /\.json$/i =~ out_file then
      out_file = out_file + ".json"
    end
    if hostname.nil? then
      hostname = Socket.gethostname
    end
    data = {
      :os => guess_os,
      :packages => get_package_informations,
      :applications => get_application_versions,
      :hostname => hostname
    }
    File.open(out_file, 'w') do |f|
      f.write(@json_builder.build(data))
    end
  end

  # アプリケーションの一覧を [ アプリ名, バージョン ]の形で
  # 配列で返す。
  def get_application_versions
    app_files = find_application_files
    analyze_application_paths(app_files)
  end

  ### 以下、対象OS/ディストリビューション毎に再定義する処理

  # OSを推測し、[ OS名, バージョン ]の形で返す。
  def guess_os
    raise "must redefine #{self.class.name}\##{__method__}."
  end

  # パッケージの一覧を配列で返す。
  def get_pakcage_informations
    raise "must redefine #{self.class.name}\##{__method__}."
  end


end
